# -*- coding: utf-8 -*-
"""
Combined output and logging facility with level support.
"""
from datetime import datetime


def runtime_error(message):
    """Stop the program using a prefixed RuntimeError."""
    output = Output()
    output.runtime_error(message)


class Output:
    """
    Combined output and logging facility with level support.
    """

    def __init__(self):
        """Initialize the logger."""
        self.debuglevel = False
        self.quiet = False

    def set_debuglevel(self, level):
        """Set the level of debugging."""
        self.debuglevel = level

    def set_quiet(self, quiet):
        """Enable or disable quiet mode."""
        self.quiet = quiet

    def runtime_error(self, message):
        """Stop the program using a prefixed RuntimeError."""
        error = ['']
        for line in message.split("\n"):
            error.append('ERROR: ' + line)
        error.append('')
        raise RuntimeError("\n".join(error))

    def log(self, message='', context='', level=0):
        """Log a string at the given indentation level."""
        if not self.debuglevel:
            return

        # Skip messages with a higher debuglevel level.
        if level+1 > self.debuglevel:
            return

        # Generate the indentation spacing.
        indent = []
        level = (level * 4) + 2
        for x in range(0, level):
            indent.append(' ')

        # Generate the prefix string.
        now = datetime.now()
        prefix = []
        prefix.append('%s  ' % now.strftime('%H:%M:%S'))
        if context:
            context = '<%s>  ' % context
            prefix.append(context.ljust(5))
        prefix.append(''.join(indent))
        prefix = ''.join(prefix)

        # Delegate output writing to write():
        if message:
            for line in message.split("\n"):
                print(prefix + line)
        else:
            print(prefix)

    def write(self, message):
        """Print the given message when not-quiet."""
        if self.quiet:
            return
        print(message)



# -*- coding: utf-8 -*-
"""
Represent a directory found on the file system.
"""
import os
from tagrenamer.fs.node import Node
from tagrenamer.fs.file import File
from tagrenamer.fs.musicfile import MusicFile, music_extensions


class Directory(Node):
    """
    Represent a directory found on the file system.

    Hooks:
    - traverse_filter (obj, path), return True/False
    - mkdir (obj)
    """

    def __init__(self, output, path, hooks={}, parent=None, dl=1):
        """Initialize the directory object."""
        self.children = []
        Node.__init__(
            self,
            output,
            path=path,
            hooks=hooks,
            parent=parent,
            dl=dl)

    def enableDryRun(self):
        """Enable a dry-run mode on this object and its children.."""
        Node.enableDryRun(self)
        if len(self.children):
            for c in self.children:
                c.enableDryRun()

    def addChild(self, child):
        """Add a child node."""
        children = []
        for c in self.children:
            children.append(c)
        children.append(child)
        self.children = children

    def removeChild(self, child):
        """Remove a child from this node."""
        children = []
        for c in self.children:
            if c.path != child.path:
                children.append(c)
        self.children = children

    def traverse(self):
        """Traverse into subdirectories and load our children."""
        self.children = []
        self.out.log(str(self), '%s.traverse' % self.type, self.dl)

        # Only really traverse if the object exists.
        if self.exists():
            for path in os.listdir(self.path):
                dl = self.dl + 1
                path = "%s/%s" % (self.path, path)

                # Invoke the traverse_filter hook (see main class description) and
                # determine if we should include this object (True) or not (False).
                filter_outcome = self.invoke('traverse_filter', self, path)
                if filter_outcome is not None:
                    if filter_outcome is False:
                        self.out.log('%s (skipping)' % str(self),
                                     '%s.traverse' % self.type,
                                     self.dl)
                        continue

                # Perform a set of tests and load the correct class for the found child.
                if os.path.isdir(path):
                    node = Directory(output=self.out,
                                     path=path,
                                     hooks=self.hooks,
                                     parent=self,
                                     dl=dl)
                    node.traverse()
                elif os.path.isfile(path):
                    extension = path.split('.').pop()
                    if extension in music_extensions:
                        node = MusicFile(output=self.out,
                                         path=path,
                                         extension=extension,
                                         hooks=self.hooks,
                                         parent=self,
                                         dl=dl)
                    else:
                        node = File(output=self.out,
                                    path=path,
                                    extension=extension,
                                    hooks=self.hooks,
                                    parent=self,
                                    dl=dl)
                else:
                    node = Node(output=self.out,
                                path=path,
                                hooks=self.hooks,
                                parent=self,
                                dl=dl)

                # Enable dry run if on the node if it applies to us.
                if self.dryrun:
                    node.enableDryRun()

                # Append the child to our list of children.
                self.children.append(node)

    def mkdir(self):
        """Make this directory if it doesn't exist on disk yet."""
        self.out.log(str(self), '%s.mkdir' % self.type, self.dl)
        if not self.dryrun:
            os.mkdir(self.path)
        self.shellCollect('mkdir -v "{}"', self.path)

        # Invoke the mkdir hook, see main class description.
        self.invoke('mkdir', self)

    def mkdirs(self, path):
        """Make multiple directories at once and assure that a path exists."""
        self.out.log(context='%s.mkdirs' % self.type, level=self.dl)

        # Calculate the sub path and current base being looked for.
        path = path.split('/')
        base = path[0]
        del path[0]

        # Determine whether the top level of the trail already exists.
        existingDir = False
        for c in self.children:
            if c.base == base:
                existingDir = c

        # Load the new directory object and create it if needed.
        if not existingDir:
            dl = self.dl + 1
            npath = "%s/%s" % (self.path, base)
            dir = Directory(output=self.out,
                            path=npath,
                            hooks=self.hooks,
                            parent=self,
                            dl=dl)
            if self.dryrun:
                dir.enableDryRun()
            dir.mkdir()
            self.children.append(dir)
        else:
            dir = existingDir

        # Let the fresh directory object recurse into itself.
        if len(path):
            return dir.mkdirs('/'.join(path))
        return dir

    def remove(self):
        """Delete this directory and it's siblings from disk."""
        self.out.log(str(self), '%s.remove' % self.type, self.dl)

        # Start with removing my children and their references.
        for c in self.children:
            c.remove()
        self.children = []

        # Remove the directory when it's emptied.
        self.shellCollect('rm -Rv "{}"', self.path)
        if not self.dryrun:
            os.rmdir(self.path)

        # Remove this instance from the parents list of children.
        if self.parent is not None:
            self.parent.removeChild(self)

        # Invoke the remove hook, see main class description.
        self.invoke('remove', self)

    def move(self, dest, onlyReferences=False):
        """Move the file system object onto a different location."""
        self.out.log(str(self), '%s.move' % self.type, self.dl)

        # Thrown an exception when we're being moved to the same location:
        if id(self) == id(dest):
            raise AssertionError("Can't move '%s' to itself" % dest)

        # Declare the new path and physically move the object.
        self.oldpath = self.path
        self.path = os.path.abspath('%s/%s' % (dest.path, self.base))
        if not onlyReferences:
            self.shellCollect('mv -v "{}" "{}"', self.oldpath, self.path)
            if not self.dryrun:
                os.rename(self.oldpath, self.path)

        # Unregister ourselves at our current parent and register at new parent.
        if self.parent:
            self.parent.removeChild(self)
            dest.addChild(self)

        # Re-parent ourselves and update several properties.
        self.parent = dest
        self.base = os.path.basename(self.path)
        self.dryrun = self.parent.dryrun
        self.root = self.parent.root
        self.relpath = self.path.replace('%s/' % self.root, '')
        self.dl = self.parent.dl + 1

        # Iterate our children and ensure they're moved too.
        for c in self.children:
            c.move(dest=self, onlyReferences=True)

        # Invoke the move hook, see main class description.
        self.invoke('move', self, dest)



# -*- coding: utf-8 -*-
"""
Represent any file, directory or other reference found on the file system.
"""
import os


class Node():
    """
    Represent any file, directory or other reference found on the file system.

    Hooks:
    - init (obj)
    - remove (obj)
    - move (obj, dest)
    - shell_collect (command)
    """
    dl = 1  # Short for debuglevel.

    def __init__(self, output, path, hooks={}, parent=None, dl=1):
        """Initialize the file-system node object."""
        self.out = output
        self.dl = self.dl + dl
        self.relpath = ''
        self.root = ''
        self.type = self.__class__.__name__
        self.path = os.path.abspath(path)
        self.base = os.path.basename(self.path)
        self.dryrun = False
        self.parent = parent
        self.hooks = hooks

        if parent is None:
            if os.path.isdir(self.path):
                self.root = self.path
            else:
                self.root = os.path.dirname(self.path)
        else:
            self.root = self.parent.root
        self.relpath = self.path.replace('%s/' % self.root, '')

        # Invoke the init hook, see main class description.
        self.invoke('init', self)

    def __str__(self):
        """Format our own base representation."""
        if ' ' in self.base:
            return "'%s'" % self.base
        return self.base

    def invoke(self, hook, *args):
        """Invoke the given hook when they have been registered at object construction."""
        if len(self.hooks) == 0:
            return
        if hook in self.hooks:
            return self.hooks[hook](*args)

    def shellCollect(self, command, *args):
        """Collect the shell equivalent of a file or directory action."""

        # Stop the call if there's no registered hook for this.
        if 'shell_collect' not in self.hooks:
            return

        # Sub function to clean incoming argument values.
        def escape(string):
            string = string.replace('"', '\\"')
            string = string.replace("&", "\&")  # noqa: W605
            string = string.replace('\/', '|')  # noqa: W605
            string = string.replace('`', '\`')  # noqa: W605
            return string

        # Create a new arguments list and escape the values.
        newargs = []
        for a in args:
            newargs.append(escape(a))
        args = tuple(newargs)

        # Parse the command and call our shell_collect hook.
        self.invoke('shell_collect', command.format(*newargs))

    def enableDryRun(self):
        """Enable a dry-run mode on this object so that no real will things happen."""
        self.dryrun = True

    def exists(self):
        """Determine whether the object really exists or not."""
        self.out.log(str(self), '%s.exists' % self.type, self.dl)
        return os.path.exists(self.path)

    def remove(self):
        """Delete the file system object from disk."""
        self.out.log(str(self), '%s.remove' % self.type, self.dl)
        self.shellCollect('rm -v "{}"', self.path)
        if not self.dryrun:
            os.unlink(self.path)

        # Remove this instance from the parents list of children.
        if self.parent is not None:
            self.parent.removeChild(self)

        # Invoke the remove hook, see main class description.
        self.invoke('remove', self)

    def move(self, dest, newFileName=None, onlyReferences=False):
        """Move the file system object onto a different location.."""
        self.out.log(str(self), '%s.move' % self.type, self.dl)

        # Thrown an exception when we're being moved to the same location:
        if id(self) == id(dest):
            raise AssertionError("Can't move '%s' to itself" % dest)

        # Declare the new path and physically move the object.
        self.oldpath = self.path
        if newFileName is not None:
            self.path = os.path.abspath('%s/%s' % (dest.path, newFileName))
        else:
            self.path = os.path.abspath('%s/%s' % (dest.path, self.base))
        if not onlyReferences:
            self.shellCollect('mv -v "{}" "{}"', self.oldpath, self.path)
            if not self.dryrun:
                os.rename(self.oldpath, self.path)

        # Log the move action for retrospection.
        self.out.log("src: '%s'" % self.oldpath.replace(self.root + '/', ''),
                     '%s.move' % self.type, self.dl + 1)
        self.out.log("dst: '%s'" % self.path.replace(self.root + '/', ''),
                     '%s.move' % self.type, self.dl + 1)

        # Unregister ourselves at our current parent and register at new parent.
        if self.parent:
            self.parent.removeChild(self)
            dest.addChild(self)

        # Re-parent ourselves and update several properties.
        self.parent = dest
        self.base = os.path.basename(self.path)
        self.dryrun = self.parent.dryrun
        self.root = self.parent.root
        self.relpath = self.path.replace('%s/' % self.root, '')
        self.dl = self.parent.dl + 1

        # Invoke the move hook, see main class description.
        self.invoke('move', self, dest)



# -*- coding: utf-8 -*-
"""
Represent the collection of all found file objects within the base path.
"""
import os
import hashlib
from tagrenamer import __version__
from tagrenamer.fs.directory import Directory
from tagrenamer.app.output import runtime_error


class Collection():
    """
    Represent the collection of all found file objects within the base path.
    """

    def __init__(self, output, settings):
        """Initialize the collection."""
        self.settings = settings
        self.out = output

        self.d_leftovers = None
        self.d_root = None
        self.d_stage = None
        self.directories = []
        self.files = []
        self.hashes = []
        self.ids = []
        self.musicfiles = []
        self.musicfiles_new_relpaths = []
        self.type = self.__class__.__name__
        self.nodes = []

        self.out.log(
            context='%s.__init__' % self.type, level=1)

        # Initialize all directory objects: the root directory,
        # the left-overs and stage directories.
        self.d_root = Directory(
            output=self.out,
            path=self.settings.dir,
            hooks={
                'init': self.callbackInit,
                'remove': self.callbackRemove,
                'move': self.callbackMove,
                'sanitize': self.callbackSanitize,
                'traverse_filter': self.callbackTraverseFilter,
                'shell_collect': self.callbackShellCollect,
                'mkdir': self.callbackMkdir})
        self.d_leftovers = Directory(
            output=self.out,
            path="%s/%s" % (self.d_root.path, self.settings.leftoversdir),
            hooks={'shell_collect': self.callbackShellCollect},
            dl=2)
        self.d_stage = Directory(
            output=self.out,
            path="%s/%s" % (self.d_root.path, self.settings.stagedir),
            hooks={'shell_collect': self.callbackShellCollect},
            parent=self.d_root,
            dl=2)

        # If we run in --dry-run mode, enable it on all directory objects.
        if self.settings.dryrun:
            self.d_root.enableDryRun()
            self.d_leftovers.enableDryRun()
            self.d_stage.enableDryRun()

    def process(self):
        """Process the collection."""
        self.out.write(" - Tagrenamer version %s." % __version__)
        self.initializeDirectories()
        self.traverse()
        self.sanitize()
        self.moveLeftovers()
        self.moveMusicToStage()
        self.removeEmptyDirectories()
        self.moveFilesPermanently()
        self.removeStageDirectory()
        self.removeLeftoversDirectory()
        self.finish()

    def __str__(self):
        """Format our own base representation."""
        return self.settings.dir.rstrip('/')

    def initializeDirectories(self):
        """Initialize the leftovers and staging directories."""
        self.out.log(
            context='%s.initializeDirectories' % self.type, level=1)
        # LEFTOVERS DIRECTORY: Clean the directory or create it.
        if self.d_leftovers.exists():
            self.d_leftovers.traverse()
            if len(self.d_leftovers.children):
                for c in self.d_leftovers.children:
                    c.remove()
            self.out.write(
                " - Leftovers directory '%s/' cleaned."
                % self.settings.leftoversdir)
        else:
            self.d_leftovers.mkdir()
            self.out.write(
                " - Leftovers directory '%s/' created."
                % self.settings.leftoversdir)

        # STAGE DIRECTORY: Create the directory or verify it is empty when it exists.
        if self.d_stage.exists():
            self.d_stage.traverse()
            if len(self.d_stage.children) != 0:
                runtime_error(
                    "Stage directory '%s/' exist but is NOT empty!"
                    % self.settings.stagedir)
            else:
                self.out.write(
                    " - Stage directory '%s/' exists."
                    % self.settings.stagedir)
        else:
            self.d_stage.mkdir()
            self.out.write(
                " - Stage directory '%s/' created." % self.settings.stagedir)

    def traverse(self):
        """Traverse the base path where the music resides in and pass our registrar."""
        self.out.log(
            context='%s.traverse' % self.type, level=1)
        self.out.write(" - Traverse the collection and extract music tags.")
        self.d_root.traverse()

    def sanitize(self):
        """Sanitize all extracted meta data for file system usage and validate input."""
        self.out.log(
            context='%s.sanitize' % self.type, level=1)
        self.out.write(" - Validating tag input and sanitizing variables.")
        for f in self.musicfiles:
            try:
                f.sanitize()
            except ValueError:
                runtime_error(
                    "Please correct the tags of this file:\n%s" % f.path)

    def moveLeftovers(self):
        """Move all the non-music files into the leftovers directory.."""
        self.out.log(
            context='%s.moveLeftovers' % self.type, level=1)
        self.out.write(
            " - Moving non music files to '%s/'." % self.settings.leftoversdir)

        # Iterate the files - which ain't music - and relocate them to the left-overs
        # directory while recreating the original directory structure. After this our
        # self.files index will be empty as they're disregarded from our index.
        for f in self.files:
            if os.path.dirname(f.relpath) != '':
                destination = self.d_leftovers.mkdirs(os.path.dirname(f.relpath))
            else:
                destination = self.d_leftovers
            f.move(destination)

    def moveMusicToStage(self):
        """Rename the music files and move them into the new structure (inside stage)."""
        self.out.log(
            context='%s.moveMusicToStage' % self.type, level=1)
        self.out.write(
            " - Moving music to new tree in stage directory '%s/'."
            % self.settings.stagedir)
        for f in self.musicfiles:
            destination_dir = os.path.dirname(f.relpath_new)
            if destination_dir == '':
                destination_dir = self.d_stage
            else:
                destination_dir = self.d_stage.mkdirs(destination_dir)
            f.move(destination_dir, os.path.basename(f.relpath_new))

    def removeEmptyDirectories(self):
        """Remove empty directories in the main music tree."""
        self.out.log(
            context='%s.removeEmptyDirectories' % self.type, level=1)
        self.out.write(
            " - Remove empty directories (except stage/leftover directories).")
        for c in self.d_root.children:
            if c.type == 'Directory':
                c.remove()

    def moveFilesPermanently(self):
        """Move all files and directories from stage to the permanent spot."""
        self.out.log(
            context='%s.moveFilesPermanently' % self.type, level=1)
        self.out.write(
            " - Move everything from stage into the final location.")
        for c in self.d_stage.children:
            c.move(self.d_root)

    def removeStageDirectory(self):
        """Remove the stage directory and object."""
        self.out.log(
            context='%s.removeStageDirectory' % self.type, level=1)
        self.out.write(
            " - Deleting the temporary stage directory '%s/'."
            % self.settings.stagedir)
        self.d_stage.remove()
        del self.d_stage
        self.d_stage = None

    def removeLeftoversDirectory(self):
        """Remove the left-overs directory and object."""
        self.out.log(
            context='%s.removeLeftoversDirectory' % self.type, level=1)
        if len(self.d_leftovers.children) == 0:
            self.out.write(
                " - Deleting the empty leftovers directory '%s/'."
                % self.settings.leftoversdir)
            self.d_leftovers.remove()
            del self.d_leftovers
            self.d_leftovers = None

    def finish(self):
        """Cleanup and drop some statistics."""
        self.out.log(
            context='%s.finish' % self.type, level=1)
        if self.settings.dryrun:
            self.out.write(
                " - DONE! Processed %d files (dry-run mode)."
                % len(self.musicfiles))
        else:
            self.out.write(
                " - DONE! Processed %d files." % len(self.musicfiles))

    # HOOK IMPLEMENTATIONS #####################################################

    def callbackInit(self, node):
        """Register a reference to any new created file system node in this collection."""
        if id(node) in self.ids:
            return
        self.ids.append(id(node))
        if node.type == 'Directory':
            self.directories.append(node)
        elif node.type == 'File':
            self.files.append(node)
        elif node.type == 'MusicFile':
            self.musicfiles.append(node)
        else:
            self.nodes.append(node)

    def callbackRemove(self, node):
        """Implementation of the remove hook - remove the object from our music index."""

        # Rewrite the ids list - without the given node.
        ids = []
        for i in self.ids:
            if i != id(node):
                ids.append(i)
        self.ids = ids

        # Rewrite any of the trees to forget the object.
        index = []
        if node.type == 'Directory':
            for d in self.directories:
                if id(d) != id(node):
                    index.append(d)
            self.directories = index
        elif node.type == 'File':
            for f in self.files:
                if id(f) != id(node):
                    index.append(f)
            self.files = index
        elif node.type == 'MusicFile':
            for f in self.musicfiles:
                if id(f) != id(node):
                    index.append(f)
            self.musicfiles = index
        else:
            for n in self.nodes:
                if id(n) != id(node):
                    index.append(n)
            self.nodes = index

        # In case of a MusicFile, lets also rewrite the hash and relpath's registry.
        if node.type == 'MusicFile':
            musicfiles_new_relpaths = []
            hashes = []
            for h in self.hashes:
                if h != node.hash_s:
                    hashes.append(h)
            for mnr in self.musicfiles_new_relpaths:
                if mnr != node.relpath_new:
                    musicfiles_new_relpaths.append(mnr)
            self.hashes = hashes
            self.musicfiles_new_relpaths = musicfiles_new_relpaths

    def callbackMove(self, node, dest):
        """Implementation of the move hook."""

        # Detect if the object being moved - for instance a left over file - goes
        # outside of our index, and forget about it if it does.
        if self.callbackMoveIsDestinationUnknown(dest):
            self.callbackRemove(node)

    def callbackMoveIsDestinationUnknown(self, dest):
        """Test if the destination or any parents aren't in our collection."""
        if id(dest) in self.ids:
            return False
        else:
            if dest.parent is not None:
                return self.callbackMoveIsDestinationUnknown(dest.parent)
            else:
                return True

    def callbackSanitize(self, node):
        """Implementation of the sanitize hook."""

        # Define a callable to generate a file based hash.
        def md5(filename, block_size=2**20):
            f = open(filename)
            md5 = hashlib.md5()
            while True:
                data = f.read(block_size)
                if not data:
                    break
                md5.update(data)
            return md5.hexdigest()

        # Verify if a song with exactly the same artist, album, title wasn't
        # submitted before and abort the process if it does.
        if node.hash_s in self.hashes:
            self.out.write("\nERROR: the following file has been identified as a duplicate!\n")
            self.out.write("What this means is that we scanned a file earlier with exactly the")
            self.out.write("same artist, album, title and extension. To prevent this from")
            self.out.write("causing any conflicts we need you to sort this out first.\n")
            self.out.write("File:   '%s'" % node.relpath)
            self.out.write("Artist: '%s'" % node.artist)
            self.out.write("Album:  '%s'" % node.album)
            self.out.write("Title:  '%s'" % node.title)
            runtime_error("Aborted, no files have been touched!")
        else:
            self.hashes.append(node.hash_s)

        # Create a dictionary with replaceable strings, our formatting arguments.
        kwarguments = {
            'artist': node.artist_s,
            'album': node.album_s,
            'title': node.title_s,
            'hash': node.hash_s,
            'ext': node.extension.lower()}

        # Parse the format and set the relpath_new field to reflect the new location.
        try:
            node.relpath_new = self.settings.format.format(**kwarguments)
        except KeyError:
            runtime_error("The provided format contains invalid fields:\n%s"
                          % self.settings.format)

        # Verify if a different file with exactly the same relpath_new isn't staged:
        if node.relpath_new in self.musicfiles_new_relpaths:
            runtime_error("The following file appears to be a duplicate!\n\n"
                          "File:     '%s'\n"
                          "New path: '%s'\n"
                          "Artist:   '%s'\n"
                          "Album:    '%s'\n"
                          "Title:    '%s'\n"
                          % (node.relpath,
                             node.relpath_new,
                             node.artist,
                             node.album,
                             node.title))
        else:
            self.musicfiles_new_relpaths.append(node.relpath_new)

    def callbackTraverseFilter(self, node, path):
        """Implementation of the traverse_filter hook."""

        # Skip including the leftovers and stage directories within the music tree.
        if path == self.d_leftovers.path:
            return False
        if path == self.d_stage.path:
            return False
        return True

    def callbackShellCollect(self, command):
        """Implementation of the shell_collect hook."""
        if self.settings.shell:
            print(command)

    def callbackMkdir(self, node):
        """Implementation of the mkdir hook."""
        pass



# -*- coding: utf-8 -*-
"""
Program arguments, defaults and Settings object.
"""
import optparse
import os
import sys
from tagrenamer import __version__
from tagrenamer.app.output import runtime_error


def get_application_context():
    """Define the program's arguments, options and default settings."""
    usage = "usage: %prog [OPTIONS] [DIRECTORY]/"
    version = "%prog {version}".format(version=__version__)
    parser = optparse.OptionParser(usage=usage, version=version)
    parser.add_option(
        "-d", "--dry-run",
        action="store_true", dest="dryrun", default=False,
        help="Perform a dry run and don't touch anything.")
    parser.add_option(
        "-f", "--format",
        action="store", type="string", dest="format", metavar='F',
        default='{artist}/{album}/{artist}-{hash}.{ext}',
        help="The format in which filenames will be rewritten.")
    parser.add_option(
        "-l", "--leftovers",
        action="store", type="string", dest="leftoversdir", metavar='L',
        default='__LEFTOVERS',
        help="The directory where non-music files will be moved to.")
    parser.add_option(
        "-S", "--stagedir",
        action="store", type="string", dest="stagedir", metavar='S',
        default='__STAGE',
        help="Temporary directory before music hits its final spot.")
    parser.add_option(
        "-s", "--shell",
        action="store_true", dest="shell", default=False,
        help="Generate and print shell commands (implies -q and -d)")
    parser.add_option(
        "-q", "--quiet",
        action="store_true", dest="quiet", default=False,
        help="Silence non-debugging output completely.")
    parser.add_option(
        "-v", "--verbose",
        action="count", dest="debuglevel", default=False, metavar='V',
        help="The level of logging verbosity.")
    (options, args) = parser.parse_args()
    return (options, args, parser)


class Settings:
    """
    Settings object.
    """

    def __init__(self):
        """Initialize the Settings object."""
        (options, args, parser) = get_application_context()
        self.options = options
        self.args = args
        self.parser = parser
        self.validate()

    def validate(self):
        """Validate input options and prevent application bugs."""

        if not len(self.args):
            self.parser.print_help()
            sys.exit(0)

        # Store the mandatory directory argument as setting key "dir".
        self.options.dir = str(self.args[0]).strip()

        # Test whether the path exists.
        if not os.path.exists(self.options.dir):
            runtime_error("Directory '%s' not found!" % self.options.dir)

        # Test whether the path is a directory as it should be.
        if not os.path.isdir(self.options.dir):
            runtime_error("Path '%s' not a directory!" % self.options.dir)

        # Enforce dry-run and quiet mode when --shell is passed.
        if self.options.shell:
            self.options.quiet = True
            self.options.dryrun = True

    def __getattr__(self, option):
        """Retrieve options via their name."""
        return self.options.__getattribute__(option)



# -*- coding: utf-8 -*-
"""
Represent a string that is absolutely filesystem safe.
"""
from unicodedata import normalize
import transliterate


class SafeString():
    """
    Represent a string that is absolutely filesystem safe.
    """
    whitelist = ('a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l',
                 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x',
                 'y', 'z', '-', ' ', '0', '1', '2', '3', '4', '5', '6', '7',
                 '8', '9')

    def __init__(self, old_string):
        """Initialize the SafeString object."""
        self.new = str(old_string)

        # Attempt to transliterate non-alphabet scripts:
        try:
            self.new = transliterate.translit(self.new, reversed=True)
        except transliterate.exceptions.LanguageDetectionError:
            pass

        # Apply common cleanup tasks:
        self.new = self.new.lower()
        self.new = self.new.strip()
        normalst = normalize("NFKD", self.new)
        self.new = ''.join(c for c in normalst if ord(c) < 0x7f)
        self.filter()

    def filter(self):
        """Filter out any garbage left.."""
        result = []
        lastWasSpace = False
        for char in self.new:
            for validchar in self.whitelist:
                if char == validchar:

                    # Prevent space duplication.
                    if char == u' ':
                        if lastWasSpace:
                            continue
                        else:
                            lastWasSpace = True
                    else:
                        lastWasSpace = False

                    # Append the character to the result list.
                    result.append(char)
        self.new = ''.join(result).strip()

    def __str__(self):
        """Return the string."""
        return self.new



# -*- coding: utf-8 -*-
"""
tagrenamer - Mass music collection renamer.
"""
import sys
from tagrenamer.app.output import runtime_error


def main():
    """Main application dispatch routine."""

    try:
        # Import the core libraries and handle import errors:
        try:
            from tagrenamer.app.output import Output
            from tagrenamer.app.settings import Settings
            from tagrenamer.app.collection import Collection
        except ImportError as e:
            runtime_error("%s\n\n"
                          "Check if these Python packages are installed:\n"
                          "- transliterate\n"
                          "- taglib (pytaglib)" % e)

        # Construct the settings and output objects:
        settings = Settings()
        output = Output()
        output.set_debuglevel(settings.debuglevel)
        output.set_quiet(settings.quiet)
        output.log(context=__name__)

        # Dispatch application control to the collection processor.
        collection = Collection(output, settings)
        collection.process()
    except RuntimeError as e:
        print(e)
        return 1
    except KeyboardInterrupt:
        print("Quitting...")
        return -1
    return 0


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
"""
Represent an ordinary file found on the file system.
"""
from tagrenamer.fs.node import Node


class File(Node):
    """
    Represent an ordinary file found on the file system.
    """

    def __init__(self, output, path, extension, hooks={}, parent=None, dl=1):
        """Initialize the file object."""
        Node.__init__(
            self,
            output=output,
            path=path,
            hooks=hooks,
            parent=parent,
            dl=dl)

        # Set the extension if available.
        self.extension = ''
        if '.' in self.base:
            self.extension = extension



# -*- coding: utf-8 -*-
"""
Represent a file of a known music file format found on the file system.
"""
import hashlib
import taglib
from tagrenamer.app.safestring import SafeString
from tagrenamer.fs.file import File


music_extensions = ('mp3', 'MP3', 'ogg', 'OGG', 'flac', 'FLAC')


class MusicFile(File):
    """
    Represent a file of a known music file format found on the file system.

    Hooks:
    - sanitize (obj)
    """

    def __init__(self, output, path, extension, hooks={}, parent=None, dl=1):
        """Initialize the music file object."""
        self.relpath_new = ''
        self.artist = ''
        self.album = ''
        self.title = ''
        self.hash = None
        self.artist_s = ''
        self.album_s = ''
        self.title_s = ''
        self.hash_s = ''
        File.__init__(
            self,
            output,
            path=path,
            extension=extension,
            hooks=hooks,
            parent=parent,
            dl=dl)

        # Set the extension if available.
        self.extension = ''
        if '.' in self.base:
            self.extension = extension

        self.extract()

    def extract(self):
        """Extract all meta data using the Taglib library."""
        self.out.log(str(self), '%s.extract' % self.type, self.dl)
        f = taglib.File(self.path)
        self.artist = ' '.join(f.tags.get('ARTIST', ['unknown_artist']))
        self.album = ' '.join(f.tags.get('ALBUM', ['unknown_album']))
        self.title = ' '.join(f.tags.get('TITLE', ['unknown_title']))
        self.out.log("Artst: '%s'" % self.artist,
                     '%s.extract' % self.type, self.dl + 1)
        self.out.log("Album: '%s'" % self.album,
                     '%s.extract' % self.type, self.dl + 1)
        self.out.log("Title: '%s'" % self.title,
                     '%s.extract' % self.type, self.dl + 1)

    def sanitize(self):
        """Sanitize the extracted data ready for file system level usage."""
        self.out.log(str(self), '%s.sanitize' % self.type, self.dl)
        self.artist_s = str(SafeString(self.artist.strip()))
        self.album_s = str(SafeString(self.album.strip()))
        self.title_s = str(SafeString(self.title.strip()))

        # Start validating the data, based on field length.
        if not len(self.artist_s):
            raise ValueError(self)
        elif not len(self.album_s):
            raise ValueError(self)
        elif not len(self.title_s):
            raise ValueError(self)

        # Generate a hash from all the strings to provide a unique file identifier.
        self.hash = hashlib.md5()
        self.hash.update(self.artist_s.encode('utf-8'))
        self.hash.update(self.album_s.encode('utf-8'))
        self.hash.update(self.title_s.encode('utf-8'))
        self.hash.update(self.extension.encode('utf-8'))
        self.hash_s = str(self.hash.hexdigest())

        # Print the sanitized metadata fields:
        self.out.log("{artist}: '%s'" % self.artist_s,
                     '%s.sanitize' % self.type, self.dl + 1)
        self.out.log("{album}: '%s'" % self.album_s,
                     '%s.sanitize' % self.type, self.dl + 1)
        self.out.log("{title}: '%s'" % self.title_s,
                     '%s.sanitize' % self.type, self.dl + 1)
        self.out.log("{hash}: '%s'" % self.hash_s,
                     '%s.sanitize' % self.type, self.dl + 1)
        self.out.log("{ext}: '%s'" % self.extension,
                     '%s.sanitize' % self.type, self.dl + 1)

        # Invoke the sanitize hook, see main class description.
        self.invoke('sanitize', self)
